// POナンバーがすでに存在するエラー
export class PoNumberAlreadyExistError extends Error {}

// 取り込むカードライセンスが存在しないエラー
export class LicenseNotExistError extends Error {}

// 取り込むライセンスが既に取り込み済みのエラー
export class LicenseKeyAlreadyActivatedError extends Error {}

// 注文不在エラー
export class OrderNotFoundError extends Error {}
// 注文発行済エラー
export class AlreadyIssuedError extends Error {}
// ライセンス不足エラー
export class LicensesShortageError extends Error {}

// ライセンス有効期限切れエラー
export class LicenseExpiredError extends Error {}
// ライセンス割り当て不可エラー
export class LicenseUnavailableError extends Error {}

// ライセンス割り当て解除済みエラー
export class LicenseAlreadyDeallocatedError extends Error {}
